@@ -6,6 +6,7 @@
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE ImpredicativeTypes #-}
 #if __GLASGOW_HASKELL__ >= 704
 {-# LANGUAGE ConstraintKinds #-}
 #endif
@@ -27,6 +28,7 @@ module Control.Monad.Trans.Resource
     , resourceForkIO
       -- * Monad transformation
     , transResourceT
+    , joinResourceT
       -- * A specific Exception transformer
     , ExceptionT (..)
     , runExceptionT_
@@ -36,6 +38,7 @@ module Control.Monad.Trans.Resource
     , allocate
     , register
     , release
+    , unprotect 
     , resourceMask
       -- * Type class/associated types
     , MonadResource (..)
@@ -47,6 +50,12 @@ module Control.Monad.Trans.Resource
     , InvalidAccess (..)
       -- * Re-exports
     , MonadBaseControl
+      -- * Internal state
+      -- $internalState
+    , InternalState
+    , getInternalState
+    , runInternalState
+    , withInternalState
     ) where
 
 import Data.Typeable
@@ -108,6 +117,7 @@ import qualified Control.Monad.ST.Lazy as LazyUnsafe
 import qualified Control.Monad.ST.Lazy as Lazy
 
 import Data.Functor.Identity (Identity, runIdentity)
+import Control.Monad.Morph
 
 -- | A lookup key for a specific release action. This value is returned by
 -- 'register' and 'allocate', and is passed to 'release'.
@@ -135,10 +145,9 @@ data ReleaseMap =
 --
 -- Since 0.3.0
 newtype ResourceT m a = ResourceT { unResourceT :: I.IORef ReleaseMap -> m a }
-
--- | Convenient alias for @ResourceT IO@.
-type ResIO a = ResourceT IO a
-
+#if __GLASGOW_HASKELL__ >= 707
+        deriving Typeable
+#else
 instance Typeable1 m => Typeable1 (ResourceT m) where
     typeOf1 = goType undefined
       where
@@ -152,6 +161,11 @@ instance Typeable1 m => Typeable1 (ResourceT m) where
 #endif
                 [ typeOf1 m
                 ]
+#endif
+
+-- | Convenient alias for @ResourceT IO@.
+type ResIO a = ResourceT IO a
+
 
 instance MonadCont m => MonadCont (ResourceT m) where
   callCC f = ResourceT $ \i -> callCC $ \c -> unResourceT (f (ResourceT . const . c)) i
@@ -208,7 +222,17 @@ register = liftResourceT . registerRIO
 --
 -- Since 0.3.0
 release :: MonadIO m => ReleaseKey -> m ()
-release (ReleaseKey istate rk) = liftIO $ release' istate rk
+release (ReleaseKey istate rk) = liftIO $ release' istate rk  
+    (maybe (return ()) id)
+
+-- | Unprotect resource from cleanup actions, this allowes you to send
+-- resource into another resourcet process and reregister it there.
+-- It returns an release action that should be run in order to clean 
+-- resource or Nothing in case if resource is already freed.
+--
+-- Since 0.4.5
+unprotect :: MonadIO m => ReleaseKey -> m (Maybe (IO ()))
+unprotect (ReleaseKey istate rk) = liftIO $ release' istate rk return
 
 -- | Perform some allocation, and automatically register a cleanup action.
 --
@@ -298,10 +322,11 @@ instance Exception InvalidAccess
 
 release' :: I.IORef ReleaseMap
          -> Int
-         -> IO ()
-release' istate key = E.mask $ \restore -> do
+         -> (Maybe (IO ()) -> IO a)
+         -> IO a
+release' istate key act = E.mask $ \restore -> do
     maction <- I.atomicModifyIORef istate lookupAction
-    maybe (return ()) restore maction
+    restore (act maction)
   where
     lookupAction rm@(ReleaseMap next rf m) =
         case IntMap.lookup key m of
@@ -366,12 +391,29 @@ bracket_ alloc cleanup inside =
 -- | Transform the monad a @ResourceT@ lives in. This is most often used to
 -- strip or add new transformers to a stack, e.g. to run a @ReaderT@.
 --
+-- Note that this function is a slight generalization of 'hoist'.
+--
 -- Since 0.3.0
 transResourceT :: (m a -> n b)
                -> ResourceT m a
                -> ResourceT n b
 transResourceT f (ResourceT mx) = ResourceT (\r -> f (mx r))
 
+-- | Since 0.4.7
+instance MFunctor ResourceT where
+    hoist f (ResourceT mx) = ResourceT (\r -> f (mx r))
+-- | Since 0.4.7
+instance MMonad ResourceT where
+    embed f m = ResourceT (\i -> unResourceT (f (unResourceT m i)) i)
+
+-- | This function mirrors @join@ at the transformer level: it will collapse
+-- two levels of @ResourceT@ into a single @ResourceT@.
+--
+-- Since 0.4.6
+joinResourceT :: ResourceT (ResourceT m) a
+              -> ResourceT m a
+joinResourceT (ResourceT f) = ResourceT $ \r -> unResourceT (f r) r
+
 -------- All of our monad et al instances
 instance Functor m => Functor (ResourceT m) where
     fmap f (ResourceT m) = ResourceT $ \r -> fmap f (m r)
@@ -412,7 +454,10 @@ instance MonadBaseControl b m => MonadBaseControl b (ResourceT m) where
      restoreM (StMT base) = ResourceT $ const $ restoreM base
 instance Monad m => MonadThrow (ExceptionT m) where
     monadThrow = ExceptionT . return . Left . E.toException
-
+instance MonadResource m => MonadResource (ExceptionT m) where
+    liftResourceT = lift . liftResourceT
+instance MonadIO m => MonadIO (ExceptionT m) where
+    liftIO = lift . liftIO
 
 -- | The express purpose of this transformer is to allow non-@IO@-based monad
 -- stacks to catch exceptions via the 'MonadThrow' typeclass.
@@ -656,3 +701,35 @@ type MonadResourceBase m = (MonadBaseControl IO m, MonadThrow m, MonadUnsafeIO m
 class (MonadBaseControl IO m, MonadThrow m, MonadUnsafeIO m, MonadIO m, Applicative m) => MonadResourceBase m
 instance (MonadBaseControl IO m, MonadThrow m, MonadUnsafeIO m, MonadIO m, Applicative m) => MonadResourceBase m
 #endif
+
+-- $internalState
+--
+-- A @ResourceT@ internally is a modified @ReaderT@ monad transformer holding
+-- onto a mutable reference to all of the release actions still remaining to be
+-- performed. If you are building up a custom application monad, it may be more
+-- efficient to embed this @ReaderT@ functionality directly in your own monad
+-- instead of wrapping around @ResourceT@ itself. This section provides you the
+-- means of doing so.
+
+-- | Get the internal state of the current @ResourceT@.
+--
+-- Since 0.4.6
+getInternalState :: Monad m => ResourceT m InternalState
+getInternalState = ResourceT return
+
+-- | The internal state held by a @ResourceT@ transformer.
+--
+-- Since 0.4.6
+type InternalState = I.IORef ReleaseMap
+
+-- | Unwrap a @ResourceT@ using the given @InternalState@.
+--
+-- Since 0.4.6
+runInternalState :: ResourceT m a -> InternalState -> m a
+runInternalState = unResourceT
+
+-- | Run an action in the underlying monad, providing it the @InternalState.
+--
+-- Since 0.4.6
+withInternalState :: (InternalState -> m a) -> ResourceT m a
+withInternalState = ResourceT
